"""PyTorch dataset for loading synchronized robot data with filesystem caching."""

import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import Callable, Optional, Set, cast

import numpy as np
import torch
from PIL import Image

import neuracore as nc
from neuracore.core.data.cache_manager import CacheManager
from neuracore.core.data.synced_dataset import SynchronizedDataset
from neuracore.core.data.synced_recording import SynchronizedRecording
from neuracore.core.nc_types import (
    CustomData,
    DataType,
    EndEffectorData,
    JointData,
    PointCloudData,
    PoseData,
    SyncPoint,
)
from neuracore.ml import BatchedTrainingSamples, MaskableData
from neuracore.ml.datasets.pytorch_neuracore_dataset import PytorchNeuracoreDataset
from neuracore.ml.utils.memory_monitor import MemoryMonitor

logger = logging.getLogger(__name__)

TrainingSample = BatchedTrainingSamples
CHECK_MEMORY_INTERVAL = 100


class PytorchSynchronizedDataset(PytorchNeuracoreDataset):
    """Dataset for loading episodic robot data from GCS with filesystem caching.

    Enhanced to support all data types including depth images, point clouds,
    poses, end-effectors, and custom sensor data.
    """

    def __init__(
        self,
        synchronized_dataset: SynchronizedDataset,
        input_data_types: list[DataType],
        output_data_types: list[DataType],
        output_prediction_horizon: int,
        cache_dir: Optional[str] = None,
        tokenize_text: Optional[
            Callable[[list[str]], tuple[torch.Tensor, torch.Tensor]]
        ] = None,
    ):
        """Initialize the dataset.

        Args:
            synchronized_dataset: The synchronized dataset to load data from.
            input_data_types: List of input data types to include in the dataset.
            output_data_types: List of output data types to include in the dataset.
            output_prediction_horizon: Number of future timesteps to predict.
            cache_dir: Directory to use for caching data.
            tokenize_text: Optional function to tokenize text data.
        """
        if not isinstance(synchronized_dataset, SynchronizedDataset):
            raise TypeError(
                "synchronized_dataset must be an instance of SynchronizedDataset"
            )
        super().__init__(
            input_data_types=input_data_types,
            output_data_types=output_data_types,
            output_prediction_horizon=output_prediction_horizon,
            tokenize_text=tokenize_text,
            num_recordings=len(synchronized_dataset),
        )
        self.synchronized_dataset = synchronized_dataset
        self.dataset_description = self.synchronized_dataset.dataset_description

        # Setup cache
        if cache_dir is None:
            cache_dir = os.path.join(tempfile.gettempdir(), "episodic_dataset_cache")
        self.cache_dir = Path(cache_dir)
        if self.cache_dir.exists():
            shutil.rmtree(self.cache_dir)  # clear cache directory if it exists
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.cache_manager = CacheManager(
            self.cache_dir,
        )

        self._max_error_count = 100
        self._error_count = 0
        self._memory_monitor = MemoryMonitor(
            max_ram_utilization=0.8, max_gpu_utilization=1.0, gpu_id=None
        )
        self._mem_check_counter = 0
        self._num_samples = self.synchronized_dataset.num_transitions
        self._logged_in = False

    @staticmethod
    def _get_timestep(episode_length: int) -> int:
        max_start = max(0, episode_length)
        return np.random.randint(0, max_start - 1)

    def load_sample(
        self, episode_idx: int, timestep: Optional[int] = None
    ) -> BatchedTrainingSamples:
        """Load sample from cache or GCS with full data type support."""
        if not self._logged_in:
            nc.login()
            self._logged_in = True
            os.environ["NEURACORE_NUM_PARALLEL_VIDEO_DOWNLOADS"] = "0"

        if self._mem_check_counter % CHECK_MEMORY_INTERVAL == 0:
            self._memory_monitor.check_memory()
            self._mem_check_counter = 0
        self._mem_check_counter += 1

        try:
            synced_recording = self.synchronized_dataset[episode_idx]
            synced_recording = cast(SynchronizedRecording, synced_recording)
            episode_length = len(synced_recording)
            timestep = timestep or self._get_timestep(episode_length)
            tensor_cache_path = self.cache_dir / f"ep_{episode_idx}_frame_{timestep}.pt"

            if tensor_cache_path.exists():
                return torch.load(tensor_cache_path, weights_only=False)
            else:
                if not self.cache_manager.ensure_space_available():
                    logger.warning("Low disk space. Some cache files were removed.")

                sample = TrainingSample(
                    output_predicition_mask=torch.ones(
                        (self.output_prediction_horizon,), dtype=torch.float32
                    ),
                )
                sync_point = cast(SyncPoint, synced_recording[timestep])
                future_sync_points = cast(
                    list[SyncPoint],
                    synced_recording[
                        timestep + 1 : timestep + 1 + self.output_prediction_horizon
                    ],
                )
                # Padding for future sync points
                for _ in range(
                    self.output_prediction_horizon - len(future_sync_points)
                ):
                    future_sync_points.append(future_sync_points[-1])

                # Process RGB images
                if sync_point.rgb_images:
                    if DataType.RGB_IMAGE in self.input_data_types:
                        rgbs_for_each_camera: list[Image.Image] = list(
                            [sp.frame for sp in sync_point.rgb_images.values()]
                        )
                        sample.inputs.rgb_images = (
                            self._create_camera_maskable_input_data(
                                rgbs_for_each_camera,
                                self.dataset_description.rgb_images.max_len,
                            )
                        )
                    if DataType.RGB_IMAGE in self.output_data_types:
                        future_frames = [
                            [cam_data.frame for cam_data in sp.rgb_images.values()]
                            for sp in future_sync_points
                            if sp.rgb_images is not None
                        ]
                        sample.outputs.rgb_images = (
                            self._create_camera_maskable_output_data(
                                future_frames,
                                self.dataset_description.rgb_images.max_len,
                            )
                        )

                # Process depth images
                if sync_point.depth_images:
                    if DataType.DEPTH_IMAGE in self.input_data_types:
                        depth_for_each_camera: list[Image.Image] = list(
                            [sp.frame for sp in sync_point.depth_images.values()]
                        )
                        sample.inputs.depth_images = (
                            self._create_camera_maskable_input_data(
                                depth_for_each_camera,
                                self.dataset_description.depth_images.max_len,
                            )
                        )
                    if DataType.DEPTH_IMAGE in self.output_data_types:
                        future_frames = [
                            [cam_data.frame for cam_data in sp.depth_images.values()]
                            for sp in future_sync_points
                            if sp.depth_images is not None
                        ]
                        sample.outputs.depth_images = (
                            self._create_camera_maskable_output_data(
                                future_frames,
                                self.dataset_description.depth_images.max_len,
                            )
                        )

                # Process point clouds
                if sync_point.point_clouds:
                    if DataType.POINT_CLOUD in self.input_data_types:
                        sample.inputs.point_clouds = (
                            self._create_point_cloud_maskable_input_data(
                                sync_point.point_clouds
                            )
                        )
                    if DataType.POINT_CLOUD in self.output_data_types:
                        future_point_clouds = [
                            sp.point_clouds
                            for sp in future_sync_points
                            if sp.point_clouds is not None
                        ]
                        sample.outputs.point_clouds = (
                            self._create_point_cloud_maskable_output_data(
                                future_point_clouds
                            )
                        )

                # Process joint data
                if sync_point.joint_positions:
                    if DataType.JOINT_POSITIONS in self.input_data_types:
                        sample.inputs.joint_positions = (
                            self._create_joint_maskable_input_data(
                                sync_point.joint_positions,
                                self.dataset_description.joint_positions.max_len,
                            )
                        )
                    if DataType.JOINT_POSITIONS in self.output_data_types:
                        sample.outputs.joint_positions = (
                            self._create_joint_maskable_output_data(
                                [
                                    sp.joint_positions
                                    for sp in future_sync_points
                                    if sp.joint_positions is not None
                                ],
                                self.dataset_description.joint_positions.max_len,
                            )
                        )

                if sync_point.joint_velocities:
                    if DataType.JOINT_VELOCITIES in self.input_data_types:
                        sample.inputs.joint_velocities = (
                            self._create_joint_maskable_input_data(
                                sync_point.joint_velocities,
                                self.dataset_description.joint_velocities.max_len,
                            )
                        )
                    if DataType.JOINT_VELOCITIES in self.output_data_types:
                        sample.outputs.joint_velocities = (
                            self._create_joint_maskable_output_data(
                                [
                                    sp.joint_velocities
                                    for sp in future_sync_points
                                    if sp.joint_velocities is not None
                                ],
                                self.dataset_description.joint_velocities.max_len,
                            )
                        )

                if sync_point.joint_torques:
                    if DataType.JOINT_TORQUES in self.input_data_types:
                        sample.inputs.joint_torques = (
                            self._create_joint_maskable_input_data(
                                sync_point.joint_torques,
                                self.dataset_description.joint_torques.max_len,
                            )
                        )
                    if DataType.JOINT_TORQUES in self.output_data_types:
                        sample.outputs.joint_torques = (
                            self._create_joint_maskable_output_data(
                                [
                                    sp.joint_torques
                                    for sp in future_sync_points
                                    if sp.joint_torques is not None
                                ],
                                self.dataset_description.joint_torques.max_len,
                            )
                        )

                if sync_point.joint_target_positions:
                    if DataType.JOINT_TARGET_POSITIONS in self.input_data_types:
                        sample.inputs.joint_target_positions = (
                            self._create_joint_maskable_input_data(
                                sync_point.joint_target_positions,
                                self.dataset_description.joint_target_positions.max_len,
                            )
                        )
                    if DataType.JOINT_TARGET_POSITIONS in self.output_data_types:
                        # We dont need to shift the sync_point by 1, since we are
                        # using the target joint positions as the action
                        jtp_points = cast(
                            list[SyncPoint],
                            synced_recording[
                                timestep : timestep + self.output_prediction_horizon
                            ],
                        )
                        for _ in range(
                            self.output_prediction_horizon - len(jtp_points)
                        ):
                            jtp_points.append(jtp_points[-1])
                        sample.outputs.joint_target_positions = (
                            self._create_joint_maskable_output_data(
                                [
                                    sp.joint_target_positions
                                    for sp in jtp_points
                                    if sp.joint_target_positions is not None
                                ],
                                self.dataset_description.joint_target_positions.max_len,
                            )
                        )

                # Process end-effector data
                if sync_point.end_effectors:
                    if DataType.END_EFFECTORS in self.input_data_types:
                        sample.inputs.end_effectors = (
                            self._create_end_effector_maskable_input_data(
                                sync_point.end_effectors
                            )
                        )
                    if DataType.END_EFFECTORS in self.output_data_types:
                        future_end_effectors = [
                            sp.end_effectors
                            for sp in future_sync_points
                            if sp.end_effectors is not None
                        ]
                        sample.outputs.end_effectors = (
                            self._create_end_effector_maskable_output_data(
                                future_end_effectors
                            )
                        )

                # Process pose data
                if sync_point.poses:
                    if DataType.POSES in self.input_data_types:
                        sample.inputs.poses = self._create_pose_maskable_input_data(
                            sync_point.poses
                        )
                    if DataType.POSES in self.output_data_types:
                        future_poses = [
                            sp.poses
                            for sp in future_sync_points
                            if sp.poses is not None
                        ]
                        sample.outputs.poses = self._create_pose_maskable_output_data(
                            future_poses
                        )

                # Process language data
                if sync_point.language_data and (
                    DataType.LANGUAGE in self.input_data_types
                    or DataType.LANGUAGE in self.output_data_types
                ):
                    if self.tokenize_text is None:
                        raise ValueError(
                            "Failed to initialize tokenize_text for DataType.LANGUAGE"
                        )
                    input_ids, attention_mask = self.tokenize_text(
                        [sync_point.language_data.text]
                    )

                    language_tokens = MaskableData(input_ids, attention_mask)
                    if DataType.LANGUAGE in self.input_data_types:
                        sample.inputs.language_tokens = language_tokens
                    if DataType.LANGUAGE in self.output_data_types:
                        sample.outputs.language_tokens = language_tokens

                # Process custom data
                if sync_point.custom_data:
                    if DataType.CUSTOM in self.input_data_types:
                        sample.inputs.custom_data = (
                            self._create_custom_maskable_input_data(
                                sync_point.custom_data
                            )
                        )
                    if DataType.CUSTOM in self.output_data_types:
                        future_custom_data = [
                            sp.custom_data
                            for sp in future_sync_points
                            if sp.custom_data is not None
                        ]
                        sample.outputs.custom_data = (
                            self._create_custom_maskable_output_data(future_custom_data)
                        )

                sample.output_predicition_mask = self._create_output_prediction_mask(
                    episode_length,
                    timestep,
                    self.output_prediction_horizon,
                )

                torch.save(sample, tensor_cache_path)

            return sample

        except Exception as e:
            import traceback

            traceback.print_exc()
            logger.error(
                f"Error loading frame {timestep} from episode {episode_idx}: {str(e)}"
            )
            raise e

    def _create_joint_maskable_input_data(
        self, joint_data: JointData, max_len: int
    ) -> MaskableData:
        """Create MaskableData for joint input."""
        jdata = torch.tensor(list(joint_data.values.values()), dtype=torch.float32)
        num_existing_states = jdata.shape[0]
        extra_states = max_len - num_existing_states
        if extra_states > 0:
            jdata = torch.cat(
                [jdata, torch.zeros(extra_states, dtype=torch.float32)], dim=0
            )
        jdata_mask = torch.tensor(
            [1.0] * num_existing_states + [0.0] * extra_states, dtype=torch.float32
        )
        return MaskableData(jdata, jdata_mask)

    def _create_joint_maskable_output_data(
        self, joint_data: list[JointData], max_len: int
    ) -> MaskableData:
        """Create MaskableData for joint output."""
        maskable_data_for_each_t = [
            self._create_joint_maskable_input_data(jd, max_len) for jd in joint_data
        ]
        stacked_maskable_data = torch.stack(
            [maskable_data.data for maskable_data in maskable_data_for_each_t]
        )
        stacked_maskable_mask = torch.stack(
            [maskable_data.mask for maskable_data in maskable_data_for_each_t]
        )
        return MaskableData(stacked_maskable_data, stacked_maskable_mask)

    def _create_end_effector_maskable_input_data(
        self, end_effector_data: EndEffectorData
    ) -> MaskableData:
        """Create MaskableData for end-effector input."""
        ee_values = list(end_effector_data.open_amounts.values())
        ee_tensor = torch.tensor(ee_values, dtype=torch.float32)

        max_len = self.dataset_description.end_effector_states.max_len
        num_existing = ee_tensor.shape[0]
        extra = max_len - num_existing

        if extra > 0:
            ee_tensor = torch.cat(
                [ee_tensor, torch.zeros(extra, dtype=torch.float32)], dim=0
            )

        ee_mask = torch.tensor(
            [1.0] * num_existing + [0.0] * extra, dtype=torch.float32
        )
        return MaskableData(ee_tensor, ee_mask)

    def _create_end_effector_maskable_output_data(
        self, end_effector_data: list[EndEffectorData]
    ) -> MaskableData:
        """Create MaskableData for end-effector output."""
        maskable_data_for_each_t = [
            self._create_end_effector_maskable_input_data(eed)
            for eed in end_effector_data
        ]
        stacked_data = torch.stack(
            [maskable_data.data for maskable_data in maskable_data_for_each_t]
        )
        stacked_mask = torch.stack(
            [maskable_data.mask for maskable_data in maskable_data_for_each_t]
        )
        return MaskableData(stacked_data, stacked_mask)

    def _create_pose_maskable_input_data(self, poses: PoseData) -> MaskableData:
        """Create MaskableData for pose input."""
        all_poses = []
        for pose_name, pose_data in poses.pose.items():
            all_poses.extend(pose_data)  # 6DOF pose

        pose_tensor = torch.tensor(all_poses, dtype=torch.float32)
        max_len = self.dataset_description.poses.max_len
        num_existing = pose_tensor.shape[0]
        extra = max_len - num_existing

        if extra > 0:
            pose_tensor = torch.cat(
                [pose_tensor, torch.zeros(extra, dtype=torch.float32)], dim=0
            )

        pose_mask = torch.tensor(
            [1.0] * num_existing + [0.0] * extra, dtype=torch.float32
        )
        return MaskableData(pose_tensor, pose_mask)

    def _create_pose_maskable_output_data(
        self, poses_list: list[PoseData]
    ) -> MaskableData:
        """Create MaskableData for pose output."""
        maskable_data_for_each_t = [
            self._create_pose_maskable_input_data(poses) for poses in poses_list
        ]
        stacked_data = torch.stack(
            [maskable_data.data for maskable_data in maskable_data_for_each_t]
        )
        stacked_mask = torch.stack(
            [maskable_data.mask for maskable_data in maskable_data_for_each_t]
        )
        return MaskableData(stacked_data, stacked_mask)

    def _create_point_cloud_maskable_input_data(
        self, point_clouds: dict[str, PointCloudData]
    ) -> MaskableData:
        """Create MaskableData for point cloud input."""
        # Stack point clouds from all sensors
        all_clouds = []
        for pc_name, pc_data in point_clouds.items():
            # Convert points to tensor: [num_points, 3]
            points = torch.tensor(pc_data.points, dtype=torch.float32)
            all_clouds.append(points)

        # For now, we'll use the first point cloud and pad to standard size
        if all_clouds:
            points = all_clouds[0]  # [num_points, 3]
            target_num_points = 1024  # Standard size
            current_num_points = points.shape[0]

            if current_num_points < target_num_points:
                # Pad with zeros
                padding = torch.zeros(target_num_points - current_num_points, 3)
                points = torch.cat([points, padding], dim=0)
            elif current_num_points > target_num_points:
                # Subsample
                indices = torch.randperm(current_num_points)[:target_num_points]
                points = points[indices]

            # Create mask for valid points
            mask = torch.tensor(
                [1.0] * min(current_num_points, target_num_points)
                + [0.0] * max(0, target_num_points - current_num_points)
            )

            # Reshape for batching: [1, num_points, 3] for single point cloud
            points = points.unsqueeze(0)
            mask = mask.unsqueeze(0)  # [1, num_points]
        else:
            # Empty point cloud
            points = torch.zeros(1, 1024, 3)
            mask = torch.zeros(1, 1024)

        return MaskableData(points, mask)

    def _create_point_cloud_maskable_output_data(
        self, point_clouds_list: list[dict[str, PointCloudData]]
    ) -> MaskableData:
        """Create MaskableData for point cloud output."""
        maskable_data_for_each_t = [
            self._create_point_cloud_maskable_input_data(pcs)
            for pcs in point_clouds_list
        ]
        stacked_data = torch.stack(
            [maskable_data.data for maskable_data in maskable_data_for_each_t]
        )
        stacked_mask = torch.stack(
            [maskable_data.mask for maskable_data in maskable_data_for_each_t]
        )
        return MaskableData(stacked_data, stacked_mask)

    def _create_custom_maskable_input_data(
        self, custom_data: dict[str, CustomData]
    ) -> dict[str, MaskableData]:
        """Create MaskableData for custom input data."""
        result = {}
        for key, data in custom_data.items():
            # Convert custom data to tensor
            if isinstance(data.data, (list, np.ndarray)):
                tensor_data = torch.tensor(data.data, dtype=torch.float32)
                if tensor_data.dim() == 0:  # Scalar
                    tensor_data = tensor_data.unsqueeze(0)

                # Create simple mask (all valid)
                mask = torch.ones(tensor_data.shape[0], dtype=torch.float32)
                result[key] = MaskableData(tensor_data, mask)
            else:
                # For other data types, create a simple representation
                tensor_data = torch.tensor(
                    [float(hash(str(data.data)) % 1000)], dtype=torch.float32
                )
                mask = torch.ones(1, dtype=torch.float32)
                result[key] = MaskableData(tensor_data, mask)

        return result

    def _create_custom_maskable_output_data(
        self, custom_data_list: list[dict[str, CustomData]]
    ) -> dict[str, MaskableData]:
        """Create MaskableData for custom output data."""
        result = {}

        # Get all keys from all timesteps
        all_keys: Set[str] = set()
        for custom_dict in custom_data_list:
            all_keys.update(custom_dict.keys())

        for key in all_keys:
            maskable_data_for_each_t = []
            for custom_dict in custom_data_list:
                if key in custom_dict:
                    single_data = self._create_custom_maskable_input_data(
                        {key: custom_dict[key]}
                    )
                    maskable_data_for_each_t.append(single_data[key])
                else:
                    # Create dummy data for missing timesteps
                    tensor_data = torch.zeros(1, dtype=torch.float32)
                    mask = torch.zeros(1, dtype=torch.float32)
                    maskable_data_for_each_t.append(MaskableData(tensor_data, mask))

            stacked_data = torch.stack(
                [maskable_data.data for maskable_data in maskable_data_for_each_t]
            )
            stacked_mask = torch.stack(
                [maskable_data.mask for maskable_data in maskable_data_for_each_t]
            )
            result[key] = MaskableData(stacked_data, stacked_mask)

        return result

    def _create_output_prediction_mask(
        self, episode_length: int, timestep: int, output_prediction_horizon: int
    ) -> torch.FloatTensor:
        """Create mask for output predictions."""
        output_prediction_mask = torch.zeros(
            output_prediction_horizon, dtype=torch.float32
        )
        for i in range(output_prediction_horizon):
            if timestep + i >= episode_length:
                break
            else:
                output_prediction_mask[i] = 1.0
        return output_prediction_mask

    def _create_camera_maskable_input_data(
        self, camera_data: list[Image.Image], max_cameras: int
    ) -> MaskableData:
        """Create MaskableData for camera input.

        Returns:
            MaskableData containing camera images of shape [num_cameras, H, W, C]
            and a mask indicating which cameras are present.
        """
        cam_image_tensors = torch.stack(
            [self.camera_transform(cam_data) for cam_data in camera_data]
        )
        num_cameras = cam_image_tensors.shape[0]
        extra_cameras = max_cameras - num_cameras
        if extra_cameras > 0:
            empty_image = torch.zeros_like(cam_image_tensors[0])
            cam_image_tensors = torch.cat(
                [cam_image_tensors, empty_image.repeat(extra_cameras, 1, 1, 1)],
                dim=0,
            )
        camera_images_mask = torch.tensor(
            [1.0] * num_cameras + [0.0] * extra_cameras,
            dtype=torch.float32,
        )
        return MaskableData(cam_image_tensors, camera_images_mask)

    def _create_camera_maskable_output_data(
        self, temporal_camera_data: list[list[Image.Image]], max_cameras: int
    ) -> MaskableData:
        """Create maskable data for multiple cameras.

        Args:
            temporal_camera_data: A list of lists of shape [T, CAMS, ...].

        Returns:
            MaskableData: A MaskableData object containing the stacked camera images
                and their masks of shape [T, CAMS, C, H, W].
        """
        maskable_data_for_each_t = [
            self._create_camera_maskable_input_data(camera_data, max_cameras)
            for camera_data in temporal_camera_data
        ]
        stacked_maskable_data = torch.stack(
            [maskable_data.data for maskable_data in maskable_data_for_each_t]
        )
        stacked_maskable_mask = torch.stack(
            [maskable_data.mask for maskable_data in maskable_data_for_each_t]
        )
        return MaskableData(stacked_maskable_data, stacked_maskable_mask)

    def __len__(self) -> int:
        """Return the number of samples in the dataset."""
        return self._num_samples



"""Diffusion Policy: Visuomotor Policy Learning via Action Diffusion."""

import logging
import time
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as T
from diffusers.schedulers.scheduling_ddim import DDIMScheduler
from diffusers.schedulers.scheduling_ddpm import DDPMScheduler

from neuracore.core.nc_types import DataType, ModelInitDescription, ModelPrediction
from neuracore.ml import (
    BatchedInferenceSamples,
    BatchedTrainingOutputs,
    BatchedTrainingSamples,
    NeuracoreModel,
)

from .modules import DiffusionConditionalUnet1d, DiffusionPolicyImageEncoder

logger = logging.getLogger(__name__)


class DiffusionPolicy(NeuracoreModel):
    """Implementation of Diffusion Policy for visuomotor policy learning.

    This implements the Diffusion Policy model for Visuomotor Policy Learning
    via Action Diffusion as described in the original paper.
    """

    def __init__(
        self,
        model_init_description: ModelInitDescription,
        hidden_dim: int = 256,
        unet_down_dims: Tuple[int, ...] = (
            512,
            1024,
            2048,
        ),
        unet_kernel_size: int = 5,
        unet_n_groups: int = 8,
        unet_diffusion_step_embed_dim: int = 128,
        unet_use_film_scale_modulation: bool = True,
        noise_scheduler_type: str = "DDPM",
        num_train_timesteps: int = 100,
        num_inference_steps: int = 100,
        beta_start: float = 0.0001,
        beta_end: float = 0.02,
        beta_schedule: str = "squaredcos_cap_v2",
        clip_sample: bool = True,
        clip_sample_range: float = 1.0,
        lr: float = 1e-4,
        lr_backbone: float = 1e-5,
        weight_decay: float = 1e-4,
        prediction_type: str = "epsilon",
    ):
        """Initialize the Diffusion Policy model.

        Args:
            model_init_description: Model initialization configuration.
            hidden_dim: Hidden dimension for image encoders.
            unet_down_dims: Downsampling dimensions for UNet.
            unet_kernel_size: Kernel size for UNet convolutions.
            unet_n_groups: Number of groups for group normalization.
            unet_diffusion_step_embed_dim: Dimension of diffusion step embeddings.
            unet_use_film_scale_modulation: Whether to use FiLM scale modulation.
            noise_scheduler_type: Type of noise scheduler ("DDPM" or "DDIM").
            num_train_timesteps: Number of timesteps for training.
            num_inference_steps: Number of timesteps for inference.
            beta_start: Starting beta value for noise schedule.
            beta_end: Ending beta value for noise schedule.
            beta_schedule: Beta schedule type.
            clip_sample: Whether to clip samples.
            clip_sample_range: Range for clipping samples.
            lr: Learning rate for main parameters.
            lr_backbone: Learning rate for backbone parameters.
            weight_decay: Weight decay for optimization.
            prediction_type: Type of prediction ("epsilon" or "sample").
        """
        super().__init__(model_init_description)
        self.lr = lr
        self.lr_backbone = lr_backbone
        self.weight_decay = weight_decay
        # Vision components
        self.image_encoders = nn.ModuleList([
            DiffusionPolicyImageEncoder(feature_dim=hidden_dim)
            for _ in range(self.dataset_description.rgb_images.max_len)
        ])
        global_cond_dim = (
            self.dataset_description.joint_positions.max_len
            + self.dataset_description.joint_velocities.max_len
            + self.dataset_description.joint_torques.max_len
        )
        if self.dataset_description.rgb_images.max_len > 0:
            global_cond_dim += (
                self.image_encoders[0].feature_dim
                * self.dataset_description.rgb_images.max_len
            )

        self.unet = DiffusionConditionalUnet1d(
            action_dim=self.dataset_description.joint_target_positions.max_len,
            global_cond_dim=global_cond_dim,
            down_dims=unet_down_dims,
            kernel_size=unet_kernel_size,
            n_groups=unet_n_groups,
            diffusion_step_embed_dim=unet_diffusion_step_embed_dim,
            use_film_scale_modulation=unet_use_film_scale_modulation,
        )

        kwargs: Dict[str, Any] = {
            "num_train_timesteps": num_train_timesteps,
            "beta_start": beta_start,
            "beta_end": beta_end,
            "beta_schedule": beta_schedule,
            "clip_sample": clip_sample,
            "clip_sample_range": clip_sample_range,
            "prediction_type": prediction_type,
        }

        self.noise_scheduler = self._make_noise_scheduler(
            noise_scheduler_type, **kwargs
        )
        self.prediction_type = prediction_type
        self.num_inference_steps = num_inference_steps
        # Normalize the images with imagenet mean and std
        self.image_normalizer = torch.nn.Sequential(
            T.Resize((224, 224)),
            T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        )

        state_mean = np.concatenate([
            self.dataset_description.joint_positions.mean,
            self.dataset_description.joint_velocities.mean,
            self.dataset_description.joint_torques.mean,
        ])
        state_std = np.concatenate([
            self.dataset_description.joint_positions.std,
            self.dataset_description.joint_velocities.std,
            self.dataset_description.joint_torques.std,
        ])
        # Register as buffers so they move with the model
        self.register_buffer(
            "joint_state_mean", self._to_torch_float_tensor(state_mean)
        )
        self.register_buffer("joint_state_std", self._to_torch_float_tensor(state_std))

        # Register as buffers so they move with the model
        self.register_buffer(
            "joint_target_mean",
            self._to_torch_float_tensor(
                self.dataset_description.joint_target_positions.mean
            ),
        )
        self.register_buffer(
            "joint_target_std",
            self._to_torch_float_tensor(
                self.dataset_description.joint_target_positions.std
            ),
        )

    def _to_torch_float_tensor(self, data: list[float]) -> torch.FloatTensor:
        """Convert list of floats to torch tensor."""
        return torch.tensor(data, dtype=torch.float32, device=self.device)

    def _preprocess_joint_state(
        self,
        joint_state: torch.FloatTensor,
        joint_state_mean: torch.FloatTensor,
        joint_state_std: torch.FloatTensor,
    ) -> torch.FloatTensor:
        """Preprocess the states."""
        return (joint_state - joint_state_mean) / joint_state_std

    def _unnormalize_actions(
        self, predicted_actions: torch.FloatTensor
    ) -> torch.FloatTensor:
        """Unnormalize the actions."""
        return (predicted_actions * self.joint_target_std) + self.joint_target_mean

    def _combine_joint_states(
        self, batch: BatchedInferenceSamples
    ) -> torch.FloatTensor:
        """Combine joint states."""
        state_inputs = []
        if batch.joint_positions:
            state_inputs.append(batch.joint_positions.data * batch.joint_positions.mask)
        if batch.joint_velocities:
            state_inputs.append(
                batch.joint_velocities.data * batch.joint_velocities.mask
            )
        if batch.joint_torques:
            state_inputs.append(batch.joint_torques.data * batch.joint_torques.mask)

        if state_inputs:
            joint_states = torch.cat(state_inputs, dim=-1)
            joint_states = self._preprocess_joint_state(
                joint_states, self.joint_state_mean, self.joint_state_std
            )
            return joint_states
        else:
            # Return zero tensor if no joint states available
            raise ValueError("No joint states available")

    def _conditional_sample(
        self,
        batch_size: int,
        prediction_horizon: int,
        global_cond: Optional[torch.Tensor] = None,
        generator: Optional[torch.Generator] = None,
    ) -> torch.Tensor:
        """Sample action sequence conditioned on the observations.

        Args:
            batch_size: Batch size
            prediction_horizon: Action sequence prediction horizon
            global_cond: Global conditioning tensor
            generator: Random number generator

        Returns:
            torch.Tensor: Sampled action sequence with shape
            (B, prediction_horizon, action_dim)
        """
        sample = torch.randn(
            size=(
                batch_size,
                prediction_horizon,
                self.dataset_description.joint_target_positions.max_len,
            ),
            dtype=torch.float32,
            device=self.device,
            generator=generator,
        )

        self.noise_scheduler.set_timesteps(self.num_inference_steps)

        for t in self.noise_scheduler.timesteps:
            # Predict model output.
            model_output = self.unet(
                sample,
                torch.full(sample.shape[:1], t, dtype=torch.long, device=sample.device),
                global_cond=global_cond,
            )
            # Compute previous image: x_t -> x_t-1
            sample = self.noise_scheduler.step(
                model_output, t, sample, generator=generator
            ).prev_sample

        return sample

    def _prepare_global_conditioning(
        self,
        joint_states: torch.FloatTensor,
        camera_images: torch.FloatTensor,
        camera_images_mask: torch.FloatTensor,
    ) -> torch.FloatTensor:
        """Encode image features and concatenate with the state vector.

        Args:
            joint_states: Joint state tensor.
            camera_images: Camera image tensor.
            camera_images_mask: Camera image mask tensor.

        Returns:
            Global conditioning tensor.
        """
        global_cond_feats = [joint_states]
        batch_size = joint_states.shape[0]
        if camera_images is not None:
            # Extract image features.
            for cam_id, encoder in enumerate(self.image_encoders):
                features = encoder(self.image_normalizer(camera_images[:, cam_id]))
                features = features * camera_images_mask[:, cam_id].view(batch_size, 1)
                global_cond_feats.append(features)

        # Concatenate features then flatten to (B, global_cond_dim).
        return torch.cat(global_cond_feats, dim=-1).flatten(start_dim=1)

    @staticmethod
    def _make_noise_scheduler(
        noise_scheduler_type: str, **kwargs: Dict[str, Any]
    ) -> Union[DDPMScheduler, DDIMScheduler]:
        """Factory for noise scheduler instances.

        All kwargs are passed to the scheduler.

        Args:
            name: Type of scheduler to create.
            **kwargs: Additional arguments for scheduler.

        Returns:
            Noise scheduler instance.
        """
        if noise_scheduler_type == "DDPM":
            return DDPMScheduler(**kwargs)
        elif noise_scheduler_type == "DDIM":
            return DDIMScheduler(**kwargs)
        else:
            raise ValueError(f"Unsupported noise scheduler type {noise_scheduler_type}")

    def _predict_action(
        self,
        batch: BatchedInferenceSamples,
        prediction_horizon: int,
    ) -> torch.Tensor:
        """Predict action sequence from observations.

        Args:
            batch: Input observations
            prediction_horizon: action sequence prediction horizon

        Returns:
            torch.FloatTensor: Predicted action sequence with shape
            (B, prediction_horizon, action_dim)
        """
        batch_size = len(batch)
        # Normalize and combine joint states
        joint_states = self._combine_joint_states(batch)

        # Encode image features and concatenate them all together along
        # with the state vector.
        if batch.rgb_images is None:
            raise ValueError("Failed to find rgb_images")
        global_cond = self._prepare_global_conditioning(
            joint_states, batch.rgb_images.data, batch.rgb_images.mask
        )  # (B, global_cond_dim)

        # run sampling
        actions = self._conditional_sample(
            batch_size, prediction_horizon, global_cond=global_cond
        )

        return actions

    def forward(self, batch: BatchedInferenceSamples) -> ModelPrediction:
        """Forward pass for inference.

        Args:
            batch: Batch of inference samples.

        Returns:
            Model prediction with outputs and timing information.
        """
        t = time.time()
        prediction_horizon = self.output_prediction_horizon
        action_preds = self._predict_action(batch, prediction_horizon)
        prediction_time = time.time() - t
        # unnormalize the actions
        predictions = self._unnormalize_actions(action_preds)
        predictions = predictions.detach().cpu().numpy()
        return ModelPrediction(
            outputs={DataType.JOINT_TARGET_POSITIONS: predictions},
            prediction_time=prediction_time,
        )

    def training_step(self, batch: BatchedTrainingSamples) -> BatchedTrainingOutputs:
        """Perform a single training step.

        Given certain timesteps, add corresponding noise to the target actions, and
        predict the added noise or the target actions, and computes mse loss.

        Args:
            batch: Training batch with inputs and targets

        Returns:
            BatchedTrainingOutputs: Training outputs with losses and metrics
        """
        inference_sample = BatchedInferenceSamples(
            joint_positions=batch.inputs.joint_positions,
            joint_velocities=batch.inputs.joint_velocities,
            joint_torques=batch.inputs.joint_torques,
            rgb_images=batch.inputs.rgb_images,
            joint_target_positions=batch.outputs.joint_target_positions,
        )
        if batch.inputs.rgb_images is None:
            raise ValueError("Failed to find rgb_images")
        joint_states = self._combine_joint_states(inference_sample)
        global_cond = self._prepare_global_conditioning(
            joint_states, batch.inputs.rgb_images.data, batch.inputs.rgb_images.mask
        )
        if batch.outputs.joint_target_positions is None:
            raise ValueError("Failed to find joint_target_positions")
        target_actions = self._preprocess_joint_state(
            batch.outputs.joint_target_positions.data,
            self.joint_target_mean,
            self.joint_target_std,
        )
        target_actions = target_actions * batch.outputs.joint_target_positions.mask
        # Sample noise to add to the trajectory.
        eps = torch.randn(target_actions.shape, device=target_actions.device)
        # Sample a random noising timestep for each item in the batch.
        timesteps = torch.randint(
            low=0,
            high=self.noise_scheduler.config.num_train_timesteps,
            size=(target_actions.shape[0],),
            device=target_actions.device,
        ).long()
        # Add noise to the clean trajectories according to the noise magnitude
        # at each timestep.
        noisy_trajectory = self.noise_scheduler.add_noise(
            target_actions, eps, timesteps
        )
        # Run the denoising network (that might denoise the trajectory, or
        # attempt to predict the noise).
        pred = self.unet(noisy_trajectory, timesteps, global_cond=global_cond)

        # Compute the loss.
        # The target is either the original trajectory, or the noise.
        if self.prediction_type == "epsilon":
            target = eps
        elif self.prediction_type == "sample":
            target = target_actions
        else:
            raise ValueError(f"Unsupported prediction type {self.prediction_type}")

        loss = F.mse_loss(pred, target, reduction="none")
        # Apply mask and reduce
        mask = batch.outputs.joint_target_positions.mask
        loss = (loss * mask).mean()

        losses = {
            "mse_loss": loss,
        }
        metrics = {
            "mse_loss": loss,
        }
        return BatchedTrainingOutputs(
            output_predicitons=pred,
            losses=losses,
            metrics=metrics,
        )

    def configure_optimizers(self) -> list[torch.optim.Optimizer]:
        """Configure optimizer with different learning rates for different components.

        Uses separate learning rates for image encoder backbone (typically lower)
        and other model parameters to account for pre-trained vision components.

        Returns:
            list[torch.optim.Optimizer]: List containing the configured optimizer
        """
        backbone_params = []
        other_params = []

        for name, param in self.named_parameters():
            if "image_encoders" in name:
                backbone_params.append(param)
            else:
                other_params.append(param)

        param_groups = [
            {"params": backbone_params, "lr": self.lr_backbone},
            {"params": other_params, "lr": self.lr},
        ]

        return [torch.optim.AdamW(param_groups, weight_decay=self.weight_decay)]

    @staticmethod
    def get_supported_input_data_types() -> list[DataType]:
        """Get the input data types supported by this model.

        Returns:
            list[DataType]: List of supported input data types
        """
        return [
            DataType.JOINT_POSITIONS,
            DataType.JOINT_VELOCITIES,
            DataType.JOINT_TORQUES,
            DataType.RGB_IMAGE,
        ]

    @staticmethod
    def get_supported_output_data_types() -> list[DataType]:
        """Get the output data types supported by this model.

        Returns:
            list[DataType]: List of supported output data types
        """
        return [DataType.JOINT_TARGET_POSITIONS]
